#!/usr/bin/env python3
import boto3
from time import sleep
from .sts import create_boto3_client


class Organization:

    def __init__(self, root_ou_id, credentials):
        """Initialise the boto3 organisation session

        :param org_client: A 'organizations' boto3 client in the Master account
        :param root_ou_id: The ID of the root Organizational Unit
        :param credentials: Tuple of (key, secret, token) of the master account
        """
        self._master_credentials = credentials
        self.root_ou_id = root_ou_id

        self._org_client = create_boto3_client(
            self.get_master_account_id(),
            'organizations',
            self._master_credentials
        )

    def list_accounts(self):
        """Retrieves all accounts in organization."""
        existing_accounts = [
            account
            for accounts in self._org_client.get_paginator("list_accounts").paginate()
            for account in accounts['Accounts']
        ]
        return existing_accounts

    def get_account_id(self, account_name):
        """Retrieves the account ID

        :param account_name: The name of the account
        :return: Id of account, None if account does not exist
        """
        for account in self.list_accounts():
            if account["Name"].strip() == account_name.strip():
                return account['Id']

        return None

    def get_ou_id(self, ou_path, parent_ou_id=None):
        """Retrieve an organisational unit id

        :param ou_path: the name of the organisational unit. For nested ou's use / notation, eg. eu/eudev
        :param parent_ou_id: The parent from where to start parsing the ou_path, defaults to root
        :return: (ou_id, parent_id)
        """
        if parent_ou_id is None:
            parent_ou_id = self.root_ou_id

        ou_hierarchy = ou_path.strip('/').split('/')
        hierarchy_index = 0

        while hierarchy_index < len(ou_hierarchy):
            response = self._org_client.list_organizational_units_for_parent(
                ParentId=parent_ou_id,
            )
            for ou in response['OrganizationalUnits']:
                if ou['Name'] == ou_hierarchy[hierarchy_index]:
                    parent_ou_id = ou['Id']
                    hierarchy_index += 1
                    break
            else:
                raise ValueError(f'Could not find ou with name {ou_hierarchy}')

        return parent_ou_id

    def move_account(self, account_id, ou_path):
        """Move the account to an organisation unit.

        Note that we are not allowing an account to move from non-root to another ou. This
        is to avoid issues with the AWS Deployment Framework that requires you to first move
        an account to root before moving to another OU.
        """
        if ou_path != self.root_ou_id:
            ou_id = self.get_ou_id(ou_path)
        else:
            ou_id = self.root_ou_id

        # TODO error handling and check if an account always returns just one item
        response = self._org_client.list_parents(ChildId=account_id)
        source_parent_id = response['Parents'][0]['Id']

        if source_parent_id == ou_id:
            # Account is already resided in ou_path
            return

        if source_parent_id != self.root_ou_id and ou_id != self.root_ou_id:
            raise ValueError(
                f"Trying to move an account from non-root OU to another OU. "
                "Move account to root ({self.root_ou_id}) first."
            )

        response = self._org_client.move_account(
            AccountId=account_id,
            SourceParentId=source_parent_id,
            DestinationParentId=ou_id
        )

    def create_account_tags(self, account_id, tags):
        """Adds tags to given account.

        :param account_id: ID of the AWS account
        :param tags: Dict of tags to apply to account
        """
        # TODO: Make the tag definition fully declarative
        # meaning that any tag not referenced will be removed
        # from the account. use org_client.untag_resource()
        formatted_tags = [
            {'Key': key, 'Value': value}
            for tag in tags
            for key, value in tag.items()
        ]
        self._org_client.tag_resource(
            ResourceId=account_id,
            Tags=formatted_tags
        )

    @staticmethod
    def get_master_account_id():
        """Retrieves the master account id from API"""
        org_client = boto3.client('organizations')
        response = org_client.describe_organization()
        return response['Organization']['MasterAccountId']

    def create_account_alias(self, account_id, account_alias):
        """Creates or updates the account alias for given account_id"""

        iam_client = create_boto3_client(
            account_id,
            "iam",
            self._master_credentials
        )

        try:
            iam_client.create_account_alias(AccountAlias=account_alias)
        except iam_client.exceptions.EntityAlreadyExistsException:
            pass  # Alias already exists

    def create_account(self, account):
        """Creates a new the account if it doesn't exist.

        :param account: Class instance of Account

        :return: account_id

            TODO: Move this to the documentation:
            # Dummy payload for account creation
                from datetime import datetime
                response = {
                    'Id': 'string',
                    'AccountName': self.full_name,
                    'State': 'SUCCEEDED',
                    'RequestedTimestamp': datetime(2015, 1, 1),
                    'CompletedTimestamp': datetime(2015, 1, 1),
                    'AccountId': '173289020951',
                    'FailureReason': 'ACCOUNT_LIMIT_EXCEEDED'
                }
            """
        if account.allow_billing is True:
            allow_billing = "ALLOW"
        else:
            allow_billing = "DENY"

        response = self._org_client.create_account(
            Email=account.email,
            AccountName=account.full_name,
            # RoleName=role_name,  # defaults to OrganizationAccountAccessRole
            IamUserAccessToBilling=allow_billing,
        )["CreateAccountStatus"]

        while response["State"] == "IN_PROGRESS":
            response = self._org_client.describe_create_account_status(
                CreateAccountRequestId=response["Id"]
            )["CreateAccountStatus"]

            if response.get("FailureReason"):
                raise IOError(
                    f"Failed to create account {account.full_name}: {response['FailureReason']}"
                )

            sleep(1)  # waiting for a sec before checking account status again

        account_id = response["AccountId"]
        # TODO: Instead of sleeping, query for the role.
        sleep(10)  # Wait until OrganizationalRole is created in new account

        return account_id


if __name__ == "__main__":
    pass



"""
# configparser.py

Module to parse and validate the yaml configuration files.
"""
import yaml
import os
from .account import Account


def read_config_files(folder):
    """Retrieve account objects from yaml configuration files in given folder.

    :param folder: Folder containing config files

    :return: list of tuples [(full_account_name, email, ou_path), ]
    """
    files = [os.path.join(folder, f) for f in os.listdir(folder)]

    accounts = []
    for filename in files:
        with open(filename, 'r') as stream:
            config = yaml.safe_load(stream)

            validate_config(config['Accounts'])

            for account in config['Accounts']:
                accounts.append(Account.load_from_config(account))

    return accounts

def validate_config(configuration):
    """Validate configuration"""
    supported_keys = [
        'OrganizationalUnitPath',
        'Email',
        'AccountFullName',
        'DeleteDefaultVPC',
        'AllowDirectMoveBetweenOU',
        'AllowBilling',
        'Alias',
        'Tags'
    ]

    if not isinstance(configuration, list):
        raise ValueError(f'Configuration invalid: {configuration}')

    for account in configuration:

        # Mandatory parameters
        if 'AccountFullName' not in account:
            raise ValueError(f'Missing AccountFullName in configuration: {account}')   
        if 'OrganizationalUnitPath' not in account:
            raise ValueError(f'Missing OrganizationalUnitPath in configuration: {account}')
        if 'Email' not in account:
            raise ValueError(f'Missing Email in configuration: {account}')

        # Optional parameters
        if 'DeleteDefaultVPC' in account and not isinstance(account['DeleteDefaultVPC'], bool):
                raise ValueError(f'{account["AccountFullName"]} DeleteDefaultVPC param should be Boolean')

        if 'AllowDirectMoveBetweenOU' in account and not isinstance(account['AllowDirectMoveBetweenOU'], bool):
                raise ValueError(f'{account["AccountFullName"]} AllowDirectMoveBetweenOU param should be Boolean')

        if 'AllowBilling' in account and not isinstance(account['AllowBilling'], bool):
                raise ValueError(f'{account["AccountFullName"]} AllowBilling param should be Boolean')

        if 'Alias' in account and not isinstance(account['Alias'], str):
                raise ValueError(f'{account["AccountFullName"]} Alias param should be String')

        if 'Tags' in account:
            if not isinstance(account['Tags'], list):
                raise ValueError(f'{account["AccountFullName"]} Tags param should be a List')
            for tag in account['Tags']:
                if not isinstance(tag, dict):
                    raise ValueError(f'{account["AccountFullName"]} Tags should be a Dict but found {tag}')

        # Invalid parameters
        for key in account:
            if key not in supported_keys:
                raise ValueError(f'Key {key} not supported in configuration: {account}')



"""
# vpc.py

Remove default VPC and related resources
"""
import logging
logger = logging.getLogger(__name__)
from time import sleep


def delete_default_vpc(client, account_id, dry_run=False):
    """Delete the default VPC in the given account id

        :param client: EC2 boto3 client instance
        :param account_id: AWS account id
        """
    # Check and remove default VPC
    default_vpc_id = None
    vpc_response = client.describe_vpcs()
    for vpc in vpc_response["Vpcs"]:
        if vpc["IsDefault"] is True:
            default_vpc_id = vpc["VpcId"]
            break

    if default_vpc_id is None:
        logging.info(f"No default VPC found in account {account_id}")
        return

    logging.info(f"Found default VPC Id {default_vpc_id}")
    subnet_response = client.describe_subnets()
    default_subnets = [
        subnet
        for subnet in subnet_response["Subnets"]
        if subnet["VpcId"] == default_vpc_id
    ]

    logging.info(f"Deleting default {len(default_subnets )} subnets")
    subnet_delete_response = [
        client.delete_subnet(SubnetId=subnet["SubnetId"], DryRun=dry_run)
        for subnet in default_subnets
    ]

    igw_response = client.describe_internet_gateways()
    try:
        default_igw = [
            igw["InternetGatewayId"]
            for igw in igw_response["InternetGateways"]
            for attachment in igw["Attachments"]
            if attachment["VpcId"] == default_vpc_id
        ][0]
    except IndexError:
        default_igw = None

    if default_igw:
        logging.info(f"Detaching Internet Gateway {default_igw}")
        detach_default_igw_response = client.detach_internet_gateway(
            InternetGatewayId=default_igw, VpcId=default_vpc_id, DryRun=dry_run
        )

        logging.info(f"Deleting Internet Gateway {default_igw}")
        delete_internet_gateway_response = client.delete_internet_gateway(
            InternetGatewayId=default_igw
        )

    sleep(10)  # It takes a bit of time for the dependencies to clear
    logging.info(f"Deleting Default VPC {default_vpc_id}")
    delete_vpc_response = client.delete_vpc(VpcId=default_vpc_id, DryRun=dry_run)

    return delete_vpc_response



"""
# sts.py

Helper fuction for working with cross-account roles
"""
import boto3
import logging


def assume_role(account_id, role_name="OrganizationAccountAccessRole", source_role=None):
    """Assume a role in a different account

    :param account_id: The AWS account ID to assume role in
    :param role_name: The name of the role to assume
    :source_role: Tuple of (access_key_id, secret_access_key, session_token) of the role you want
                                to assume from

    :return: Tuple of (access_key_id, secret_access_key, session_token)
    """
    role_arn = f"arn:aws:iam::{account_id}:role/{role_name}"

    if source_role is None:
        sts_client = boto3.client("sts")
    else:
        sts_client = boto3.client(
            "sts",
            aws_access_key_id=source_role[0],
            aws_secret_access_key=source_role[1],
            aws_session_token=source_role[2]
        )

    current_account = sts_client.get_caller_identity().get('Account')
    logging.debug(f"Account {current_account} assuming role {role_name} on account {account_id}")

    sts_response = sts_client.assume_role(
        RoleArn=role_arn,
        RoleSessionName='newsession'
    )

    return (
        sts_response["Credentials"]["AccessKeyId"],
        sts_response["Credentials"]["SecretAccessKey"],
        sts_response["Credentials"]["SessionToken"]
    )

def create_boto3_client(account_id, client_type, source_role_credentials=None, region_name=None):
        """Creates a boto3 client associated with the account_id you want to target.

        The client is created by first assuming a role in the master account and from there
        assume the OrganizationAccountAccessRole in a member account. This is done to get
        this code working through ADF running in the deployment account.

        :param account_id: The account_id you want to create an account in
        :param client_type: The type of boto3 client you want to create, eg. 'sts' or 'ec2'
        :param source_role_credentials: The assumed role to assume the new session from
        :param region_name: The region to intialize in
        """
        # Assume OrganizationAccountAccessRole through Master account
        if source_role_credentials is None:
            logging.debug(f'Creating client {client_type}.')
            access_key_id, secret_access_key, session_token = assume_role(account_id)
        else:
            logging.debug(f'Creating client {client_type} through a source_role.')
            access_key_id, secret_access_key, session_token = assume_role(
                account_id,
                source_role=source_role_credentials
            )

        client = boto3.client(
            client_type,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token,
            region_name=region_name
        )
        return client



import setuptools

with open('README.md', 'r') as f:
    readme = f.read()

with open('HISTORY.md', 'r') as f:
    history = f.read()

setuptools.setup(
    name="awsaccountmgr",
    version="0.0.4",
    author="Mathijs Mortimer",
    keywords="AWS Accounts",
    author_email="mathijs@mortimer.nl",
    description="A command line tool for managing accounts within an AWS organization. Easy to integrate into AWS Deployment Framework",
    long_description=readme + '\n\n' + history,
    long_description_content_type="text/markdown",
    url="https://github.com/thiezn/awsaccountmgr/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    scripts=['bin/awsaccountmgr'],
)


"""
# account.py
Model for handling AWS accounts within the organization. The Account class
allows you to create or update a new account.
"""

from time import sleep


class Account:
    def __init__(
        self,
        full_name,
        email,
        ou_path,
        alias=None,
        delete_default_vpc=False,
        allow_direct_move_between_ou=False,
        allow_billing=True,
        tags=None
    ):
        """Initialize Account object.

        :param full_name: Full name of the account
        :param email: Valid email address of account
        :param ou_path: Organizational Unit path. For nested OUs use / notation, eg. us/development.
                        Use root identifier to target root OU (eg. r-abc1)

        :param alias: Optional Alias to use, if None the alias will default to full_name
        :param delete_default_vpc: Set this to True to delete the default vpc from the account
        :param allow_direct_move_between_ou: Set this to False to prevent the script from moving
                                             an account directly from one OU to another. This is
                                             useful if you are using the AWS Deployment Framework
                                             as it requires you to first move an account to the root
                                             to trigger the base cloudformation stack updates.
        :param allow_billing: Set this to False to prevent account admins from using the billing console
        :param tags: a dict containing optional tags for the account
        """
        self.full_name = full_name
        self.email = email
        self.ou_path = ou_path
        self.delete_default_vpc = delete_default_vpc
        self.allow_direct_move_between_ou = allow_direct_move_between_ou
        self.allow_billing = allow_billing

        if alias is None:
            self.alias = full_name
        else:
            self.alias = alias

        if tags is None:
            self.tags = {}
        else:
            self.tags = tags

    @classmethod
    def load_from_config(cls, config):
        """Initialize Account class from configuration object"""
        return cls(
            config["AccountFullName"],
            config["Email"],
            config["OrganizationalUnitPath"],
            alias=config.get("Alias", None),
            delete_default_vpc=config.get("DeleteDefaultVPC", False),
            allow_direct_move_between_ou=config.get("AllowDirectMoveBetweenOU", False),
            allow_billing=config.get("AllowBilling", True),
            tags=config.get("Tags", {})
        )



from .configparser import read_config_files, validate_config
from .sts import assume_role, create_boto3_client
from .vpc import delete_default_vpc
from .account import Account
from .organization import Organization